;***************************************************************************
; Clock V2.0     (c) ASIX s.r.o. 1997
; Source code for assembler compiler:
;   MPASM 01.21 Released, (c) 1993-95 Byte Craft Limited
; Output Hex file can be programmed with
;   PVKpro V1.0  (c) ASIX s.r.o. 1997
;***************************************************************************
        list    F=inhx8m, p=16c84, r=hex
        __config        3ff1h                   ; fuses
; External oscillator frequency: 3.2768MHz => instruction cycle = 1.22s
; 1.22s x 256 (8-bit tmr0) x 16 (prescaler ratio) = 5ms (tmr0 period).
; Fuses: Osc: XT, PuT: On, WDT: Off, CP: Off.
;***************************************************************************
; Pins:
;   RA0 common anode - the forth  digit from the left (DISP0)      (out)
;   RA1 common anode - the thirth digit from the left (DISP1)      (out)
;   RA2 common anode - the second digit from the left (DISP2)      (out)
;   RA3 common anode - the first  digit from the left (DISP3)      (out)
;   RA4 common anode - LED diodes (D0 - D7)                        (out)
;   RB0 - RB7 - seven segments and decimal point (multiplexed)     (out)
;   RB0 button TL0 - start / stop                                  (in)
;   RB1 button TL1 - display mode switching                        (in)
;   RB2 button TL2 - hour setting                                  (in)
;   RB3 button TL3 - minute setting                                (in)
;***************************************************************************
; predefined registers of processor
indf    equ     00h     ; f0 register
tmr0    equ     01h     ; real time clock/counter register
pcl     equ     02h     ; low byte of program counter
status  equ     03h     ; status word register
fsr     equ     04h     ; file select register
porta   equ     05h     ; port A
portb   equ     06h     ; port B
pclath  equ     0ah     ; program counter latch high
intcon  equ     0bh     ; interrupt sources register
roption equ     01h     ; option register
trisa   equ     05h     ; trisa register
trisb   equ     06h     ; trisb register

; definition of user's registers
sec     equ     10h     ; seconds in BCD
min     equ     11h     ; minutes in BCD
hour    equ     12h     ; hours in BCD
                        ; this order has to be respected

counter equ     13h     ; counter 1 - seconds
endr    equ     14h     ; register to store the final value after setting
stat1   equ     15h     ; system state
                        ; bit0: 0 - no button pushed (initial state)
                        ;       1 - the button reckognized to be pushed
                        ;           and still held
                        ; bit1: 0 - the clock is running (initial state)
                        ;       1 - the clock is halted
                        ; bit2: 0 - hours and minutes are displayed
                        ;           (initial state)
                        ;       1 - minutes and seconds are displayed
count_m equ     16h     ; multiplex counter
                        ; it counts modulo 4: 0,1,2,3,0,1,2,3,0,1,...
digit   equ     17h     ; register for the digit displayed
new     equ     18h     ; register for the state of the buttons - new value
old     equ     19h     ; register for the state of the buttons - old value
count_d equ     1ah     ; time interval counter

; definition of constants
w       equ     00h     ; destination is the working register
f       equ     01h     ; destination is the specified file register

; definition of substitutions
#define carry   status, 0       ; more comprehensible then single C
#define dcarry  status, 1       ; more comprehensible then single DC
#define zero    status, 2       ; more comprehensible then single Z
#define push    stat1, 0        ; stat1 register - bit 0
#define stop    stat1, 1        ; stat1 register - bit 1
#define hm      stat1, 2        ; stat1 register - bit 2
#define dp      portb, 7        ; port B - bit 7 (decimal point)

; definition of macroinstructions
bnk0    macro                   ; macroinstruction for selection bank 0
        bcf     status, 6
        bcf     status, 5
        endm
bnk1    macro                   ; macroinstruction for selection bank 1
        bcf     status, 6
        bsf     status, 5
        endm

        org     000h            ; the program starts from address 000h
;***************************************************************************
        clrf    pclath          ; reset pclath before executing jumps and
                                ; subroutines
        goto    init            ; jump to beginning of the program
;***************************************************************************
; The routine transforms BCD digits to seven segment display code. It uses
; the w register.
; Input: BCD digit in w.
; Output: the value in w to be sent to port B to display the appropriate
; digit. The segment is on, if the bit on port B is in log 0.
tab1    addwf   pcl, f          ; digit selection
        retlw   b'11000000'     ; 0 is displayed
        retlw   b'11111001'     ; 1 is displayed
        retlw   b'10100100'     ; 2 is displayed
        retlw   b'10110000'     ; 3 is displayed
        retlw   b'10011001'     ; 4 is displayed
        retlw   b'10010010'     ; 5 is displayed
        retlw   b'10000010'     ; 6 is displayed
        retlw   b'11111000'     ; 7 is displayed
        retlw   b'10000000'     ; 8 is displayed
        retlw   b'10010000'     ; 9 is displayed
;***************************************************************************
; The routine sets the active anode. It depends on count_m. Both input and
; output are in the w register as well. This time the output is sent to the
; port A. The anode is active if the bit on port A is in log 0.
tab2    addwf   pcl, f          ; anode selection
        retlw   b'11111110'     ; count_m=0 - anode 0 is active
        retlw   b'11111101'     ; count_m=1 - anode 1 is active
        retlw   b'11111011'     ; count_m=2 - anode 2 is active
        retlw   b'11110111'     ; count_m=3 - anode 3 is active
;***************************************************************************
; The routine displays one of the four digits. The actual digit is selected
; from the hm bit of the stat1 register and from the value of the count_m
; (multiplex counter).
display movlw   0ffh
        movwf   porta           ; disconnection of all anodes
        movlw   sec
        movwf   fsr             ; address for the sec register to fsr
        btfss   hm              ; are hours and minutes displayed? (bit=0?)
        incf    fsr, f          ; yes => address of min register is in fsr
                                ; no  => nothing is happening
        bcf     carry           ; reset of the carry bit (for rotation)
        rrf     count_m, w      ; multiplex counter divided by 2 is in w
                                ; carry = 1, if count_m was odd
                                ; carry = 0, if count_m was even
        iorlw   00h             ; conditionally sets bit 2 of the status
                                ; it is necessary, the rrf instruction does
                                ; not affect it
        btfss   zero            ; initial value of count_m = 0 or 1? (Z=1?)
        incf    fsr, f          ; no  => fsr points to the next register:
                                ;        min - in case of mode min, sec
                                ;        hour - in case of mode hour, min
                                ; yes => nothing is happening
        movf    indf, w         ; selected register to w
        movwf   digit           ; w to digit
        btfsc   carry           ; was count_m odd? (C=1?)
        swapf   digit, f        ; yes => swap nibbles
                                ; no  => leave unchanged
        movlw   b'00001111'
        andwf   digit, w        ; masking
                                ; the result is one BCD digit
        call    tab1            ; BCD digit to seven segment code
        movwf   portb           ; segment codes to port B
        bcf     carry           ; reset of the carry bit (for rotation)
        rrf     sec, w          ; seconds divided by 2 to w
        btfsc   carry           ; is the even second displayed? (C=0?)
        goto    disp_a          ; no  => decimal point is off
                                ; yes => decimal point depends on count_m
        movf    count_m, w
        xorlw   02h
        btfsc   zero            ; count_m = 2? (Z=1?)
        bcf     dp              ; yes => decimal point is on
                                ; no  => nothing is happening
disp_a  movf    count_m, w      ; count_m to w
        call    tab2            ; anode selection (with respect to count_m)
        movwf   porta           ; activation of the anode
        incf    count_m, f      ; incrementation of the count_m
        movlw   b'00000011'
        andwf   count_m, f      ; masking
                                ; count_m counts to 3 only in this case
        retlw   00h             ; end
;***************************************************************************
; The routine debuggs oscillations on buttons and sets the appropriate bits
; of stat1 register, which control the state of the system. The routine reads
; buttons every 20ms (using count_m register) and if the button is pushed
; at least for two readings immediately followed one after another, the button
; is reckognized to be pushed.
button  movf    count_m, w
        btfss   zero            ; count_m = 0? (Z=1?)
        retlw   00h             ; no  => end
        movlw   b'11111111'     ; yes => reading of buttons
        movwf   porta           ; disconnecting of anodes
        movwf   portb           ; disconnecting of catodes
        bnk1                    ; switch to bank 1
        movlw   b'11111111'
        movwf   trisb           ; port B is an input again
        bnk0                    ; return to bank 0
        movlw   .56
        movwf   count_d         ; time interval counter = 56
butt_a  decfsz  count_d, f      ; time delay to stabilize the state of the bus
        goto    butt_a          ; about 200s
        movf    portb, w        ; reading of port B (the state of the buttons)
                                ; the button is pushed => the pin is in log 0
        xorlw   b'11111111'     ; complement of the state of the buttons
        andlw   b'00001111'     ; masking of not significant bits
                                ; the button is pushed => the bit of the
                                ; register is in log 1
        movwf   new             ; rewriting of the state of the buttons to
                                ; new register
        movlw   b'11111111'     ; disconnecting of catodes before switching
        movwf   portb           ; of the direction
        bnk1                    ; switch to bank 1
        movlw   b'00000000'
        movwf   trisb           ; port B is an output again
        bnk0                    ; return to bank 0
        btfss   push            ; is the button pushed yet? (bit push=1?)
        goto    butt_b          ; no  => continue
                                ; yes => test for releasing the button
        clrf    old             ; reset of the old register
        movf    new, w
        btfsc   zero            ; new = 0? (Z=1?)
        bcf     push            ; yes => the button has been released, the bit
                                ;        has been cleared, end
        retlw   00h             ; no  => nothing but the end

butt_b  movf    old, w          ; the old state of the buttons to w
        andwf   new, w          ; w contains new and old values
        btfss   zero            ; new and old values = 0? (Z=1?)
        goto    butt_c          ; no  => the button is pushed
        movf    new, w          ; yes => rewriting of the state of the
                                ;        buttons to old register
        movwf   old
        retlw   00h             ; end

butt_c  bsf     push            ; the button is held, setting of the bit
        btfss   new, 0          ; is the button start/stop pushed? (bit0=1?)
        goto    butt_d          ; no  => continue
        movlw   b'00000010'     ; yes => complement of the stop bit,
                                ;        other bits unchanged
        xorwf   stat1, f
        btfsc   stop            ; is the clock halted? (stop bit = 1?)
        clrf    sec             ; yes => reset of seconds and then the end
        retlw   00h             ; no  => nothing but the end

butt_d  btfss   new, 1          ; is the button switch pushed? (bit1=1?)
        goto    butt_e          ; no  => continue
        movlw   b'00000100'     ; yes => complement of the hm bit,
                                ;        other bits unchanged
        xorwf   stat1, f
        retlw   00h             ; end

butt_e  btfss   new, 2          ; is the button hour setting pushed? (bit2=1?)
        goto    butt_f          ; no  => the button minute setting is pushed
        movlw   hour            ; yes => increment of hours
        movwf   fsr
        movlw   24h
        movwf   endr            ; selection of hours
        call    rewrite         ; rewriting of hours
        retlw   00h             ; end

butt_f  movlw   min             ; increment of minutes
        movwf   fsr
        movlw   60h
        movwf   endr            ; selection of minutes
        call    rewrite         ; rewriting of minutes
        retlw   00h             ; end
;***************************************************************************
; The rewriting routine. Input is an address of the register to be rewritten
; (stored in fsr beforehand) and the value for overflow (the endr register).
; The routine increments (with BCD correction) the register pointed by fsr.
; If this register (after the correction) equals to the endr register, the
; register is cleared and the zero flag is set.
rewrite movlw   .7
        addwf   indf, f         ; increments indf with BCD correction
        btfsc   dcarry
        goto    rew_a
        movlw   .6
        subwf   indf, f
rew_a   movf    endr, w
        xorwf   indf, w
        btfsc   zero            ; indf = endr? (Z=1?)
        clrf    indf            ; yes => reset of indf
        retlw   00h             ; no  => end
;***************************************************************************
; The routine waits for 5ms (not since calling the routine, but since last
; tmr0 overflow).
_5ms    comf    tmr0, w
        btfss   zero            ; tmr0 = 0ffh? (Z=1?)
        goto    _5ms            ; no  => wait
                                ; yes => test for overflow
_5ms_a  btfsc   tmr0, 0         ; has tmr0 overflowed? (bit0=0?)
        goto    _5ms_a          ; no  => wait for overflow
        retlw   00h             ; yes => end
;***************************************************************************
; The routine, which initializes ports and the state of the system, resets
; hours, minutes, seconds and sets the mode of the time counting and
; displaying of hours and minutes.
init    bnk0                    ; switch to bank 0
        clrf    intcon          ; interrupts are disabled
        movlw   0ffh
        movwf   porta           ; disconnecting of anodes
        movwf   portb           ; disconnecting of catodes
        bnk1                    ; switch to bank 1
        movlw   00h
        movwf   trisa           ; port A as an output
        movwf   trisb           ; port B as an output
        movlw   b'10000011'     ; tmr0 internally driven, prescaler 1:16
        movwf   roption         ; pull-up disable
        bnk0                    ; return to bank 0
        clrf    stat1           ; initial setting of the system
        clrf    count_m         ; reset of the multiplex counter
        clrf    hour            ; reset of hours
        clrf    min             ; reset of minutes
        clrf    sec             ; reset of seconds
        clrf    new             ; reset of the state of the buttons - new
        clrf    old             ; reset of the state of the buttons - old
        clrf    tmr0            ; reset of tmr0
        goto    real            ; end, jump to time counting and displaying
;***************************************************************************
real    movlw   .200
        movwf   counter         ; the main counter for 1s
real_a  decfsz  counter, f      ; is 1s passed?
        goto    real_b          ; no  => continue
        movlw   .200
        movwf   counter         ; yes => initialization of the counter
        btfsc   stop            ; is the clock running? (bit stop=0?)
        goto    real_b          ; no  => continue
        movlw   sec             ; yes => new time
        movwf   fsr
        movlw   60h
        movwf   endr            ; selection of seconds
        call    rewrite         ; setting of seconds
        btfss   zero            ; sec = 60? (Z=1?)
        goto    real_b          ; no  => continue
        movlw   min             
        movwf   fsr             ; yes => selection of minutes
        call    rewrite         ; setting of minutes
        btfss   zero            ; min = 60? (Z=1?)
        goto    real_b          ; no  => continue
        movlw   hour
        movwf   fsr
        movlw   24h
        movwf   endr            ; yes => selection of hours
        call    rewrite         ; setting of hours
real_b  call    button          ; routine for reading the buttons
        call    display         ; time displaying
        call    _5ms            ; wait for 5ms
        goto    real_a          ; and so on ...
;***************************************************************************
        end
